from async_sqs_consumer.queue import (
    Queue,
)
from async_sqs_consumer.utils import (
    TASK_NAME_PREFIX,
    validate_message,
)
from async_sqs_consumer.utils.retry import (
    retry_call,
)
import asyncio
from contextlib import (
    suppress,
)
from datetime import (
    datetime,
)
from functools import (
    partial,
)
import json
import logging
import pytz
import signal
import sys
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
)
from uuid import (
    uuid4,
)

logging.basicConfig(level=logging.INFO)

LOGGER = logging.getLogger(__name__)


class Worker:  # pylint: disable=too-many-instance-attributes
    def __init__(
        self, queues: dict[str, Queue], max_workers: Optional[int] = None
    ) -> None:
        self.queues = queues
        self.running = False
        self.handlers: dict[str, Callable[..., Any]] = {}
        self.handlers_by_queue: dict[str, dict[str, Callable[..., Any]]] = {}
        self._events: dict[str, list[Callable[[], None]]] = {}
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._tasks: list[asyncio.Task] = []
        self._max_workers = max_workers or 1024

    def _sigint_handler(self, *_args: Any) -> None:
        sys.stdout.write("\b\b\r")
        sys.stdout.flush()
        logging.getLogger("system").warning(
            "Received <ctrl+c> interrupt [SIGINT]"
        )
        self._stop_worker()

    def _sigterm_handler(self, *_args: Any) -> None:
        logging.getLogger("system").warning(
            "Received termination signal [SIGTERM]"
        )
        self._stop_worker()

    def _stop_worker(self, *_: Any) -> None:
        if self._loop:
            LOGGER.debug("Stop worker")
            self.running = False
            for queue in self.queues.values():
                queue.stop_polling()
            for event in self._events.get("shutdown", []):
                self._loop.run_until_complete(event())  # type: ignore
            for task in self._tasks:
                task.cancel()
            _queue_tasks = [
                task
                for task in asyncio.all_tasks(self._loop)
                if task.get_name().startswith(TASK_NAME_PREFIX)
            ]
            # TODO: make sure all tasks are finished before finishing the entire process

            self._loop.stop()

    def task(
        self, name: str, queue_name: Optional[str] = None
    ) -> Callable[..., None]:
        """A decorator to add a task that could be handled by the worker.

        Args:
            name (str): tasks received with this name will be
            processed by the decorated function
        Returns:
            Callable[[Callable[[], None]], None]: _description_
        """

        def decorator(func: Callable[[], Any]) -> None:
            if queue_name:
                if queue_name not in self.handlers_by_queue:
                    self.handlers_by_queue[queue_name] = {name: func}
                else:
                    self.handlers_by_queue[queue_name][name] = func
            else:
                self.handlers[name] = func

        return decorator

    def on_event(
        self, event_name: Literal["startup", "shutdown"]
    ) -> Callable[[Callable[[], None]], None]:
        """Yo can define event handlers that need to be executed before
        the application startups, or when the application is shutting down.
        The valid events are `startup`, `shutdown`.

        Args:
            event_name (Literal[&quot;startup&quot;, &quot;shutdown&quot;]):

        Returns:
            Callable[[Callable[[], None]], None]:
        """

        def decorator(func: Callable[[], None]) -> None:
            if event_name not in self._events:
                self._events[event_name] = [func]
            else:
                self._events[event_name] = [*self._events[event_name], func]

        return decorator

    def start(
        self, event_loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> None:
        """Start the workert

        Args:
            event_loop (Optional[asyncio.AbstractEventLoop], optional):
            if an eventloop is already running you can run the worker in it.
            Defaults to None.
        """
        self.running = True
        loop = event_loop or asyncio.get_event_loop()
        self._loop = loop

        if self._loop and self._loop.is_closed():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        for signame in ("SIGINT", "SIGTERM"):
            self._loop.add_signal_handler(
                getattr(signal, signame), partial(self._stop_worker, self)
            )

        signal.siginterrupt(signal.SIGTERM, False)
        signal.siginterrupt(signal.SIGUSR1, False)
        signal.signal(signal.SIGINT, partial(self._sigint_handler, self))
        signal.signal(signal.SIGTERM, partial(self._sigterm_handler, self))

        for queue_name, queue in self.queues.items():
            polling_daemon_task = loop.create_task(
                queue.start_polling(
                    self._consumer_callback,
                    queue_name,
                    max_parallel_messages=self._max_workers,
                )
            )
            self._tasks = [*self._tasks, polling_daemon_task]

        for event in self._events.get("startup", []):
            self._loop.run_until_complete(event())  # type: ignore

        self._loop.run_forever()

    def _delete_message(
        self,
        _future: Optional[asyncio.Future],
        *,
        handler_name: str,
        task_id: str,
        receipt_handle: Any,
        queue_alias: str,
        start_time: float,
    ) -> None:
        start_date = datetime.fromtimestamp(start_time, pytz.UTC)
        end_date = datetime.utcnow().replace(tzinfo=pytz.UTC)
        total_seconds = end_date - start_date
        LOGGER.info(
            "Task %s[%s] succeeded in %ss: None",
            handler_name,
            task_id,
            total_seconds.seconds,
        )
        asyncio.ensure_future(
            self.queues[queue_alias].delete_messages(receipt_handle),
            loop=self._loop,
        )

    async def _process_message(
        self, message_content: dict[str, Any], queue_alias: str
    ) -> bool:
        try:
            body = json.loads(message_content["Body"])
        except json.JSONDecodeError:
            return False

        if not validate_message(body):
            LOGGER.error("Failed to validate message")
            return False

        LOGGER.info("Task %s[%s] received", body["task"], body["id"])
        handler = self.handlers_by_queue.get(queue_alias, {}).get(body["task"])
        handler = handler or self.handlers.get(body["task"])
        if not handler:
            LOGGER.warning(
                "The task %s has not a candidate to be proceeded",
                body["task"],
            )
            return False

        if not self._loop:
            LOGGER.warning("The main event loop is not initialized")
            return False
        task = self._loop.create_task(
            retry_call(
                handler,
                fargs=body.get("args", []),
                fkwargs=body.get("kwargs", {}),
                tries=body.get("retries", 1),
            ),
            name=(
                f"{TASK_NAME_PREFIX}{queue_alias}"
                f"_{body['task']}_{uuid4().hex[:8]}"
            ),
        )
        task.add_done_callback(
            partial(
                self._delete_message,
                receipt_handle=message_content["ReceiptHandle"],
                queue_alias=queue_alias,
                start_time=datetime.utcnow().timestamp(),
                handler_name=body["task"],
                task_id=body["id"],
            )
        )
        return True

    async def _consumer_callback(
        self, message: dict[str, Any], queue_alias: str
    ) -> None:
        with suppress(asyncio.CancelledError):
            success = await self._process_message(message, queue_alias)
            if not success:
                await self.queues[queue_alias].delete_messages(
                    message["ReceiptHandle"]
                )



from typing import (
    NamedTuple,
    Optional,
)

MESSAGE_SCHEMA = {
    "type": "object",
    "properties": {
        "task": {"type": "string"},
        "id": {"type": "string"},
        "args": {"type": "array"},
        "kwargs": {"type": "object"},
        "retries": {"type": "number"},
        "eta": {"type": "string"},
        "expires": {"type": "string"},
    },
    "required": ["task", "id"],
}


class AwsCredentials(NamedTuple):
    access_key_id: str
    secret_access_key: str
    session_token: Optional[str] = None


class SqsOptions(NamedTuple):
    visibility_timeout: Optional[int] = None


class Context(NamedTuple):
    queue_url: str
    aws_credentials: Optional[AwsCredentials] = None
    sqs: Optional[SqsOptions] = None



import aioboto3
from aiobotocore.config import (
    AioConfig,
)
from async_sqs_consumer.types import (
    AwsCredentials,
)
from contextlib import (
    AsyncExitStack,
)
from typing import (
    Any,
    Optional,
)

RESOURCE_OPTIONS_SQS = {
    "aws_access_key_id": (None),
    "aws_secret_access_key": (None),
    "config": AioConfig(
        # The time in seconds till a timeout exception is thrown when
        # attempting to make a connection. [60]
        connect_timeout=15,
        # Maximum amount of simultaneously opened connections. [10]
        # https://docs.aiohttp.org/en/stable/client_advanced.html#limiting-connection-pool-size
        max_pool_connections=2000,
        # The time in seconds till a timeout exception is thrown when
        # attempting to read from a connection. [60]
        read_timeout=30,
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html
        retries={"max_attempts": 10, "mode": "standard"},
        # Signature version for signing URLs
        # https://boto3.amazonaws.com/v1/documentation/api/1.9.42/guide/s3.html#generating-presigned-urls
        signature_version="s3v4",
    ),
    "endpoint_url": None,
    "region_name": "us-east-1",
    "service_name": "sqs",
    "use_ssl": True,
    "verify": True,
}

SESSION = aioboto3.Session()
CONTEXT_STACK_SQS = None
RESOURCE_SQS = None


async def sqs_startup(credentials: Optional[AwsCredentials] = None) -> None:
    # pylint: disable=global-statement
    global CONTEXT_STACK_SQS, RESOURCE_SQS

    CONTEXT_STACK_SQS = AsyncExitStack()
    if credentials:
        RESOURCE_OPTIONS_SQS["aws_access_key_id"] = credentials.access_key_id
        RESOURCE_OPTIONS_SQS[
            "aws_secret_access_key"
        ] = credentials.secret_access_key
        RESOURCE_OPTIONS_SQS["aws_session_token"] = credentials.session_token

    RESOURCE_SQS = await CONTEXT_STACK_SQS.enter_async_context(
        SESSION.client(**RESOURCE_OPTIONS_SQS)
    )


async def sqs_shutdown() -> None:
    if CONTEXT_STACK_SQS:
        await CONTEXT_STACK_SQS.aclose()


async def get_sqs_client(credentials: Optional[AwsCredentials] = None) -> Any:
    if RESOURCE_SQS is None:
        await sqs_startup(credentials)

    return RESOURCE_SQS



from async_sqs_consumer.queue import (
    Queue,
)
from async_sqs_consumer.worker import (
    Worker,
)
from asyncio import (
    sleep,
)

worker = Worker(
    queues={
        "default": Queue(
            url="https://sqs.us-east-1.amazonaws.com/205810638802/integrates_tasks_dev"
        )
    },
    max_workers=10,
)


@worker.task("report", queue_name="default")
async def report(text: str) -> None:
    print(text)
    await sleep(5)
    print("finish task")


if __name__ == "__main__":
    worker.start()



from async_sqs_consumer.types import (
    MESSAGE_SCHEMA,
)
from contextlib import (
    suppress,
)
from jsonschema import (
    validate,
)
from jsonschema.exceptions import (
    ValidationError,
)
from typing import (
    Any,
)

TASK_NAME_PREFIX = "task_queue_"


class FailedValidation(Exception):
    pass


def validate_message(message: dict[str, Any]) -> bool:
    with suppress(ValidationError):
        validate(message, MESSAGE_SCHEMA)
        return True
    return False



from aiohttp.client_exceptions import (
    ClientConnectorError,
    ServerDisconnectedError,
)
from async_sqs_consumer.resources import (
    get_sqs_client,
    RESOURCE_OPTIONS_SQS,
    SESSION,
)
from async_sqs_consumer.types import (
    AwsCredentials,
)
from async_sqs_consumer.utils import (
    TASK_NAME_PREFIX,
)
from async_sqs_consumer.utils.retry import (
    retry,
)
import asyncio
from asyncio import (
    get_event_loop,
    sleep,
)
from botocore.client import (
    BaseClient,
)
from botocore.exceptions import (
    ClientError,
)
from contextlib import (
    suppress,
)
import logging
from typing import (
    Any,
    Callable,
    Coroutine,
    Optional,
)

LOGGER = logging.getLogger(__name__)
NETWORK_ERRORS = (
    ServerDisconnectedError,
    ClientConnectorError,
    ClientError,
    asyncio.TimeoutError,
)


@retry(exceptions=NETWORK_ERRORS, tries=3, delay=0.2)
async def get_queue_messages(
    queue_url: str,
    credentials: Optional[AwsCredentials] = None,
    client: Optional[BaseClient] = None,
    visibility_timeout: Optional[int] = None,
) -> list[dict[str, Any]]:
    client = await get_sqs_client(credentials)
    response = await client.receive_message(
        QueueUrl=queue_url,
        MaxNumberOfMessages=10,
        VisibilityTimeout=visibility_timeout or 60,
    )
    return response.get("Messages", [])


@retry(exceptions=NETWORK_ERRORS, tries=3, delay=0.2)
async def delete_messages(
    queue_url: str,
    receipt_handle: dict[str, Any],
    credentials: Optional[AwsCredentials] = None,
) -> None:
    client = await get_sqs_client(credentials)
    await client.delete_message(
        ReceiptHandle=receipt_handle,
        QueueUrl=queue_url,
    )


class Queue:
    def __init__(  # pytlint: disable=too-many-arguments,too-many-arguments
        self,
        url: str,
        authentication: Optional[AwsCredentials] = None,
        polling_interval: Optional[float] = None,
        visibility_timeout: Optional[int] = None,
        max_queue_parallel_messages: Optional[int] = None,
    ) -> None:
        self.url = url
        self.authentication = authentication
        self.polling_interval = polling_interval or 1.0
        self.visibility_timeout = visibility_timeout or 60
        self._max_queue_parallel_messages = max_queue_parallel_messages
        self._polling = False

    async def start_polling(
        self,
        callback: Callable[[dict[str, Any], str], Coroutine[Any, Any, None]],
        queue_alias: str,
        max_parallel_messages: Optional[int] = None,
    ) -> None:
        self._polling = self._polling or True
        async with SESSION.client(**RESOURCE_OPTIONS_SQS) as sqs_client:
            while self._polling:
                if len(
                    [
                        task
                        for task in asyncio.all_tasks(get_event_loop())
                        if task.get_name().startswith(TASK_NAME_PREFIX)
                    ]
                ) > (
                    self._max_queue_parallel_messages
                    or max_parallel_messages
                    or 1024
                ):
                    await sleep(self.polling_interval)
                    continue

                with suppress(asyncio.CancelledError):
                    messages = await get_queue_messages(
                        self.url,
                        client=sqs_client,
                        visibility_timeout=self.visibility_timeout,
                    )
                    await asyncio.gather(
                        *[
                            callback(message, queue_alias)
                            for message in messages
                        ]
                    )
                    await sleep(self.polling_interval)

    def stop_polling(self) -> None:
        self._polling = False

    async def delete_messages(
        self,
        receipt_handle: dict[str, Any],
    ) -> None:
        async with SESSION.client(**RESOURCE_OPTIONS_SQS) as sqs_client:
            await sqs_client.delete_message(
                ReceiptHandle=receipt_handle,
                QueueUrl=self.url,
            )



import asyncio
import functools
from functools import (
    partial,
)
import logging
import random
import traceback
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
)

logging_logger = logging.getLogger(__name__)


async def __retry_internal(  # pylint: disable=too-many-arguments
    func: partial,
    tries: int,
    exceptions: Union[
        Type[BaseException], list[type[BaseException]]
    ] = Exception,
    delay: Optional[float] = None,
    max_delay: Optional[float] = None,
    backoff: Optional[float] = None,
    jitter: Union[Literal[1], Literal[0]] = 0,
    logger: Optional[logging.Logger] = None,
    log_traceback: bool = False,
    on_exception: Optional[Callable[[Type[Exception]], bool]] = None,
):
    _tries, _delay = tries, (delay or 0)
    logger = logger or logging_logger
    while _tries:
        try:
            return await func()
        except exceptions as exc:
            if on_exception is not None:
                if on_exception(exc):
                    break

            _tries -= 1
            if not _tries:
                raise

            if logger is not None:
                try:
                    func_qualname = func.func.__qualname__
                except AttributeError:
                    func_qualname = str(func.func)
                logger.warning(
                    "%s: %s in %s.%s, retrying in %s seconds...",
                    exc.__class__.__qualname__,
                    exc,
                    func.func.__module__,
                    func_qualname,
                    _delay,
                )
                if log_traceback:
                    logger.warning(traceback.format_exc())

            await asyncio.sleep(_delay)
            _delay *= backoff or 1

            if isinstance(jitter, tuple):
                _delay += random.uniform(*jitter)
            else:
                _delay += jitter

            if max_delay is not None:
                _delay = min(_delay, max_delay)


def retry(
    tries: int,
    exceptions: Union[
        Type[BaseException], list[type[BaseException]]
    ] = Exception,
    delay: Optional[float] = None,
    max_delay: Optional[float] = None,
    backoff: Optional[float] = None,
    jitter: Union[Literal[1], Literal[0]] = 0,
    logger: Optional[logging.Logger] = None,
    log_traceback: bool = False,
    on_exception: Optional[Callable[[Type[Exception]], bool]] = None,
):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*fargs, **fkwargs):
            args = fargs if fargs else []
            kwargs = fkwargs if fkwargs else {}
            return await __retry_internal(
                partial(func, *args, **kwargs),
                tries,
                exceptions,
                delay,
                max_delay,
                backoff,
                jitter,
                logger,
                log_traceback,
                on_exception,
            )

        return wrapper

    return decorator


async def retry_call(
    func: partial,
    tries: int,
    exceptions: Union[
        Type[BaseException], list[type[BaseException]]
    ] = Exception,
    delay: Optional[float] = None,
    max_delay: Optional[float] = None,
    backoff: Optional[float] = None,
    jitter: Union[Literal[1], Literal[0]] = 0,
    logger: Optional[logging.Logger] = None,
    fargs: Optional[Tuple[Any, ...]] = None,
    fkwargs: Optional[dict[str, Any]] = None,
):
    args = fargs if fargs else tuple()
    kwargs = fkwargs if fkwargs else {}
    return await __retry_internal(
        partial(func, *args, **kwargs),
        tries,
        exceptions,
        delay,
        max_delay,
        backoff,
        jitter,
        logger,
    )



import boto3
import json
import uuid

client = boto3.client("sqs", region_name="us-east-1")

for _ in range(1):
    client.send_message(
        QueueUrl=(
            "https://sqs.us-east-1.amazonaws.com/205810638802/integrates_tasks_dev"
        ),
        MessageBody=json.dumps(
            {
                "id": uuid.uuid4().hex,
                "task": "clone",
                "args": [
                    "vergeladc",
                    "ddafdd56-c17f-43c3-9e4c-7e3e074399b9",
                ],
            }
        ),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['async_sqs_consumer', 'async_sqs_consumer.test', 'async_sqs_consumer.utils']

package_data = \
{'': ['*']}

install_requires = \
['aioboto3>=10.1.0,<11.0.0',
 'aiobotocore>=2.4.0,<3.0.0',
 'jsonschema>=4.17.3,<5.0.0',
 'pyrate-limiter>=2.10.0,<3.0.0']

setup_kwargs = {
    'name': 'async-sqs-consumer',
    'version': '0.3.2',
    'description': '',
    'long_description': '# Async SQS consumer\n\nPython asynchronous (**async** / **await**) worker for consuming messages\nfrom AWS SQS.\n\nThis is a hobby project, if you find the project interesting\nany contribution is welcome.\n\n## Usage\n\nYou must create an instance of the worker with the url of the queue.\n\nAws credentials are taken from environment variables, you must set the\nfollowing environment variables. Or you can provide a Context object with the\naws credentials `async_sqs_consumer.types.Context`\n\n- `AWS_ACCESS_KEY_ID`\n- `AWS_SECRET_ACCESS_KEY`\n\nExample:\n\nYou can get the queue url with the follow aws cli command\n`aws sqs get-queue-url --queue-name xxxxxx`\n\n```python\n# test_worker.py\n\nfrom async_sqs_consumer.worker import (\n    Worker,\n)\n\nworker = Worker(\n    queue_url="https://sqs.us-east-1.amazonaws.com/xxxxxxx/queue_name"\n)\n\n\n@worker.task("report")\nasync def report(text: str) -> None:\n    print(text)\n\nif __name__: "__main__":\n    worker.start()\n```\n\nNow you can initialize the worker `python test_worker.py`\n\nNow you need to post a message for the worker to process\n\n```python\nimport json\nimport boto3\nimport uuid\n\nclient = boto3.client("sqs")\n\nclient.send_message(\n    QueueUrl="https://sqs.us-east-1.amazonaws.com/xxxxxxx/queue_name",\n    MessageBody=json.dumps(\n        {\n            "task": "report",\n            "id": uuid.uuid4().hex,\n            "args": ["hello world"],\n        }\n    ),\n)\n```\n\nOr you can use aioboto3\n\n```python\nimport asyncio\nimport json\nimport aioboto3\nimport uuid\n\n\nasync def main() -> None:\n    session = aioboto3.Session()\n    async with session.client("sqs") as client:\n        await client.send_message(\n            QueueUrl="https://sqs.us-east-1.amazonaws.com/xxxxxxx/queue_name",\n            MessageBody=json.dumps(\n                {\n                    "task": "report",\n                    "id": uuid.uuid4().hex,\n                    "args": ["hello world"],\n                }\n            ),\n        )\n\n\nif __name__ == "__main__":\n    asyncio.run(main())\n```\n\nTo publish the messages they must have the following structure\n\n```json\n{\n    "type": "object",\n    "properties": {\n        "task": {"type": "string"},\n        "id": {"type": "string"},\n        "args": {"type": "array"},\n        "kwargs": {"type": "object"},\n        "retries": {"type": "number"},\n        "eta": {"type": "string"},\n        "expires": {"type": "string"},\n    },\n    "required": ["task", "id"],\n}\n```\n',
    'author': 'Diego Restrepo',
    'author_email': 'drestrepo@fluidattacks.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/drestrepom/async_sqs_consumer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
